==> cryptide-1.5.2.tgz: package/src/elGamalKey.js <==
// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import assert from 'assert';
import BigInt from 'big-integer';
import createHash from 'create-hash';
import secretShare from './secretShare.js';
import utils, { constants } from './utils.js';
import Hash from './hash.js';
import PHelp from './phelp.js';

export default class ElGamalKey {
    /**
     * 
     * @param {boolean} privated 
     * @param {bigInt.BigInteger} p
     * @param {bigInt.BigInteger} q
     * @param {bigInt.BigInteger} g
     * @param {bigInt.BigInteger} key
     * @param {bigInt.BigInteger} id
     */
    constructor(privated, p, q, g, key, id = undefined) {
        this.privated = privated;
        this.p = p;
        this.q = q;
        this.g = g;
        this.key = key; //represent the secret in elgammal and the y in secret sharing
        this.id = id; //represent the x value in secret sharing
    }

    get size() {
        return this.p.bitLength().valueOf();
    }

    createPublic() {
        const key = this.g.modPow(this.key, this.p);
        return new ElGamalKey(false, this.p, this.q, this.g, key);
    }

    share(number, threshold) {
        const [xs, ys] = secretShare.share(this.key, number, threshold, this.q);
        return ys.map((yi, i) => new ElGamalKey(this.privated, this.p, this.q, this.g, yi, xs[i]));
    }

    static combine(shares) {
        const first = shares[0];
        const [ids, keys] = [shares.map(sh => sh.id), shares.map(sh => sh.key)]
        const key = secretShare.interpolate(ids, keys, first.q);

        return new ElGamalKey(first.privated, first.p, first.q, first.g, key)
    }

    toString() {
        const values = [this.p, this.g, this.key];
        if (this.id)
            values.push(this.id);

        const bytes = utils.encodeByteArray(values, this.size, true);
        var checksum = Hash.shaBuffer(bytes)[0];
        const falgs = ((this.privated ? 1 : 0) << 1 | (this.id ? 1 : 0)) << 6 | (checksum >> 2);

        var size = Math.ceil(this.size / 8);
        bytes.unshift(size & 0xff);
        bytes.unshift((size >> 8) & 0xff);
        bytes.unshift(falgs);
        
        return Buffer.from(bytes).toString('base64');
    }

    static from(text) {
        const [flags, s0, s1, ...bytes] = Buffer.from(text, 'base64');
        const privated = (flags & 0b10000000) > 0;
        const numbers = (flags & 0b1000000) ? 4 : 3;
        var checksum = Hash.shaBuffer(bytes)[0];
        var size = (s0 << 8) | s1

        assert(bytes.length == numbers * size && checksum >> 2 === (flags & 63) , 'Malformed Key');

        const [p, g, key, id ] = utils.decode(bytes, bytes.length * 8 / numbers);
        const q = p.minus(constants.one).divide(constants.two);

        return new ElGamalKey(privated, p, q, g, key, id);
    }

    static generate(bits = 256, confidence = 32) {
        const [p, q] = utils.randomPrime(bits, confidence);
        const g = utils.getPrimitiveRoot(p).modPow(constants.two, p);
        
        const x = utils.random(constants.one, q.subtract(constants.one));
        const y = g.modPow(x, p);

        return [new ElGamalKey(true, p, q, g, x), new ElGamalKey(false, p, q, g, y)];
    }

    /**
     * @returns {Promise<ElGamalKey[]>}
     */
    static generateAsync(bits = 256, confidence = 32) {
        return PHelp.run(() => ElGamalKey.generate(bits, confidence));
    }

    /**
     * @param {string|Uint8Array} key
     */
    static seed(key) {
        const [p, q, g] = getSecureParams();

        const bytes = p.bitLength().valueOf() / 8;
        const x = seedPrivate(key, bytes);
        const y = g.modPow(x, p);

        return [new ElGamalKey(true, p, q, g, x), new ElGamalKey(false, p, q, g, y)];
    }
}

/**
 * @param {string|Uint8Array} key
 * @param {number} size
 */
function seedPrivate(key, size) {
    var bytes = [...Array(Math.ceil(size / 64))].map(() => (key = Hash.sha512Buffer(key)));
    return BigInt.fromArray(Array.from(Buffer.concat(bytes).slice(0, size)), 256, false);
}

function getSecureParams() {
    var p = BigInt("0FFFFFFFFFFFFFFFFC90FDAA22168C234C4C6628B80DC1CD129024E088A67CC74020BBEA63B139B22514A08798E3404DDEF9519B3CD3A431B302B0A6DF25F14374FE1356D6D51C245E485B576625E7EC6F44C42E9A637ED6B0BFF5CB6F406B7EDEE386BFB5A899FA5AE9F24117C4B1FE649286651ECE45B3DC2007CB8A163BF0598DA48361C55D39A69163FA8FD24CF5F83655D23DCA3AD961C62F356208552BB9ED529077096966D670C354E4ABC9804F1746C08CA18217C32905E462E36CE3BE39E772C180E86039B2783A2EC07A28FB5C55DF06F4C52C9DE2BCBF6955817183995497CEA956AE515D2261898FA051015728E5A8AACAA68FFFFFFFFFFFFFFFF", 16);
    var g = BigInt("35dce0815f01e2f940976518465cfb2c01c0e1ea24a42b2ad09a50e421054fafe57eb30e04db92fa27956351b338da2a759a91bb0b5fb11478f970c1f444756c6d1353d2fa3ebba95489bb5c6874dfce8cc426acb3f577376a261d43dffff699a7caedb9869a69cd770cee2cf4f42cd40623f323e1a938a1d63ff4b44c3b465dc131bf15a48c1caa4e07cfa258bb994b8481b93fa45b774a2cd1214b8516919d8afed78bec3b5108677b07bc3ad67d9d4007bdb0e0959b0874f3ab0c53817a665c5f058a84d0a7ea782d7b11f693252574c90579a962ec4042d7287044b44c10488b5552ebabad62c70c860c65fec9230a0ff2392b8f9366fbc8c7f72aedeac9", 16);
    const q = p.minus(BigInt.one).shiftRight(1);

    return [p, q, g];
}

==> cryptide-2.0.14.tgz: package/src/elGamalKey.js <==
// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import assert from 'assert';
import BigInt from 'big-integer';
import secretShare from './secretShare.js';
import utils, { constants } from './utils.js';
import Hash from './hash.js';
import PHelp from './phelp.js';


export let params = class Params {
    static get p() { return BigInt("0FFFFFFFFFFFFFFFFC90FDAA22168C234C4C6628B80DC1CD129024E088A67CC74020BBEA63B139B22514A08798E3404DDEF9519B3CD3A431B302B0A6DF25F14374FE1356D6D51C245E485B576625E7EC6F44C42E9A637ED6B0BFF5CB6F406B7EDEE386BFB5A899FA5AE9F24117C4B1FE649286651ECE45B3DC2007CB8A163BF0598DA48361C55D39A69163FA8FD24CF5F83655D23DCA3AD961C62F356208552BB9ED529077096966D670C354E4ABC9804F1746C08CA18217C32905E462E36CE3BE39E772C180E86039B2783A2EC07A28FB5C55DF06F4C52C9DE2BCBF6955817183995497CEA956AE515D2261898FA051015728E5A8AACAA68FFFFFFFFFFFFFFFF", 16); }
    static get g() { return BigInt("35dce0815f01e2f940976518465cfb2c01c0e1ea24a42b2ad09a50e421054fafe57eb30e04db92fa27956351b338da2a759a91bb0b5fb11478f970c1f444756c6d1353d2fa3ebba95489bb5c6874dfce8cc426acb3f577376a261d43dffff699a7caedb9869a69cd770cee2cf4f42cd40623f323e1a938a1d63ff4b44c3b465dc131bf15a48c1caa4e07cfa258bb994b8481b93fa45b774a2cd1214b8516919d8afed78bec3b5108677b07bc3ad67d9d4007bdb0e0959b0874f3ab0c53817a665c5f058a84d0a7ea782d7b11f693252574c90579a962ec4042d7287044b44c10488b5552ebabad62c70c860c65fec9230a0ff2392b8f9366fbc8c7f72aedeac9", 16); }
    static get q() { return Params.p.minus(BigInt.one).shiftRight(1); }
};

export default class ElGamalKey {
    /**
     * 
     * @param {boolean} privated 
     * @param {bigInt.BigInteger} p
     * @param {bigInt.BigInteger} q
     * @param {bigInt.BigInteger} g
     * @param {bigInt.BigInteger} key
     * @param {bigInt.BigInteger} id
     */
    constructor(privated, p, q, g, key, id = undefined) {
        this.privated = privated;
        this.p = p;
        this.q = q;
        this.g = g;
        this.key = key; //represent the secret in elgammal and the y in secret sharing
        this.id = id; //represent the x value in secret sharing
    }

    get size() {
        return this.p.bitLength().valueOf();
    }

    createPublic() {
        const key = this.g.modPow(this.key, this.p);
        return new ElGamalKey(false, this.p, this.q, this.g, key);
    }

    share(number, threshold) {
        const [xs, ys] = secretShare.share(this.key, number, threshold, this.q);
        return ys.map((yi, i) => new ElGamalKey(this.privated, this.p, this.q, this.g, yi, xs[i]));
    }

    static combine(shares) {
        const first = shares[0];
        const [ids, keys] = [shares.map(sh => sh.id), shares.map(sh => sh.key)]
        const key = secretShare.interpolate(ids, keys, first.q);

        return new ElGamalKey(first.privated, first.p, first.q, first.g, key)
    }

    toString() {
        const values = [this.p, this.g, this.key];
        if (this.id)
            values.push(this.id);

        const bytes = utils.encodeByteArray(values, this.size, true);
        var checksum = Hash.shaBuffer(bytes)[0];
        const falgs = ((this.privated ? 1 : 0) << 1 | (this.id ? 1 : 0)) << 6 | (checksum >> 2);

        var size = Math.ceil(this.size / 8);
        bytes.unshift(size & 0xff);
        bytes.unshift((size >> 8) & 0xff);
        bytes.unshift(falgs);
        
        return Buffer.from(bytes).toString('base64');
    }

    static from(text) {
        const [flags, s0, s1, ...bytes] = Buffer.from(text, 'base64');
        const privated = (flags & 0b10000000) > 0;
        const numbers = (flags & 0b1000000) ? 4 : 3;
        var checksum = Hash.shaBuffer(bytes)[0];
        var size = (s0 << 8) | s1

        assert(bytes.length == numbers * size && checksum >> 2 === (flags & 63) , 'Malformed Key');

        const [p, g, key, id ] = utils.decode(bytes, bytes.length * 8 / numbers);
        const q = p.minus(constants.one).divide(constants.two);

        return new ElGamalKey(privated, p, q, g, key, id);
    }

    static generate(bits = 256, confidence = 32) {
        const [p, q] = utils.randomPrime(bits, confidence);
        const g = utils.getPrimitiveRoot(p).modPow(constants.two, p);
        
        const x = utils.random(constants.one, q.subtract(constants.one));
        const y = g.modPow(x, p);

        return [new ElGamalKey(true, p, q, g, x), new ElGamalKey(false, p, q, g, y)];
    }

    /**
     * @returns {Promise<ElGamalKey[]>}
     */
    static generateAsync(bits = 256, confidence = 32) {
        return PHelp.run(() => ElGamalKey.generate(bits, confidence));
    }

    /**
     * @param {string|Uint8Array} key
     */
    static seed(key) {
        const [p, q, g] = [params.p, params.q, params.g]

        const bytes = p.bitLength().valueOf() / 8;
        const x = seedPrivate(key, bytes);
        const y = g.modPow(x, p);

        return [new ElGamalKey(true, p, q, g, x), new ElGamalKey(false, p, q, g, y)];
    }
}

/**
 * @param {string|Uint8Array} key
 * @param {number} size
 */
function seedPrivate(key, size) {
    var bytes = [...Array(Math.ceil(size / 64))].map(() => (key = Hash.sha512Buffer(key)));
    return BigInt.fromArray(Array.from(Buffer.concat(bytes).slice(0, size)), 256, false);
}
